@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2011 Alon Bar-Lev <alon.barlev@gmail.com>
+ * Copyright (c) 2005-2018 Alon Bar-Lev <alon.barlev@gmail.com>
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -114,9 +114,7 @@ __pkcs11h_threading_atfork_child (void);
 #endif
 static
 CK_RV
-__pkcs11h_forkFixup (
-	IN const PKCS11H_BOOL activate_slotevent
-);
+__pkcs11h_forkFixup ();
 #endif
 
 
@@ -524,7 +522,7 @@ CK_RV
 pkcs11h_setForkMode (
 	IN const PKCS11H_BOOL safe
 ) {
-#if defined(ENABLE_PKCS11H_THREADING) && !defined(_WIN32)
+#if !defined(_WIN32)
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 
@@ -653,9 +651,6 @@ pkcs11h_addProvider (
 	IN const unsigned slot_poll_interval,
 	IN const PKCS11H_BOOL cert_is_private
 ) {
-#if defined(ENABLE_PKCS11H_THREADING)
-	PKCS11H_BOOL mutex_locked = FALSE;
-#endif
 #if defined(ENABLE_PKCS11H_DEBUG)
 #if defined(_WIN32)
 	int mypid = 0;
@@ -698,13 +693,6 @@ pkcs11h_addProvider (
 		provider_location
 	);
 
-#if defined(ENABLE_PKCS11H_THREADING)
-	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
-		goto cleanup;
-	}
-	mutex_locked = TRUE;
-#endif
-
 	if ((rv = _pkcs11h_mem_malloc ((void *)&provider, sizeof (struct _pkcs11h_provider_s))) != CKR_OK) {
 		goto cleanup;
 	}
@@ -806,6 +794,12 @@ pkcs11h_addProvider (
 
 	provider->enabled = TRUE;
 
+#if defined(ENABLE_PKCS11H_THREADING)
+	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
+		goto cleanup;
+	}
+#endif
+
 	if (_g_pkcs11h_data->providers == NULL) {
 		_g_pkcs11h_data->providers = provider;
 	}
@@ -821,6 +815,11 @@ pkcs11h_addProvider (
 	}
 
 	provider = NULL;
+
+#if defined(ENABLE_PKCS11H_THREADING)
+	_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
+#endif
+
 	rv = CKR_OK;
 
 cleanup:
@@ -839,13 +838,6 @@ cleanup:
 		provider = NULL;
 	}
 
-#if defined(ENABLE_PKCS11H_THREADING)
-	if (mutex_locked) {
-		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
-		mutex_locked = FALSE;
-	}
-#endif
-
 #if defined(ENABLE_PKCS11H_SLOTEVENT)
 	_pkcs11h_slotevent_notify ();
 #endif
@@ -877,6 +869,7 @@ pkcs11h_removeProvider (
 	PKCS11H_BOOL has_mutex_global = FALSE;
 	PKCS11H_BOOL has_mutex_cache = FALSE;
 	PKCS11H_BOOL has_mutex_session = FALSE;
+	CK_RV lock_rv;
 #endif
 	_pkcs11h_provider_t provider = NULL;
 	CK_RV rv = CKR_FUNCTION_FAILED;
@@ -896,16 +889,18 @@ pkcs11h_removeProvider (
 	);
 
 #if defined(ENABLE_PKCS11H_THREADING)
-	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.cache)) != CKR_OK) {
-		goto cleanup;
+	lock_rv = CKR_OK;
+
+	if ((lock_rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.cache)) != CKR_OK) {
+		goto free1;
 	}
 	has_mutex_cache = TRUE;
-	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.session)) != CKR_OK) {
-		goto cleanup;
+	if ((lock_rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.session)) != CKR_OK) {
+		goto free1;
 	}
 	has_mutex_session = TRUE;
-	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
-		goto cleanup;
+	if ((lock_rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
+		goto free1;
 	}
 	has_mutex_global = TRUE;
 
@@ -926,12 +921,44 @@ pkcs11h_removeProvider (
 		provider = provider->next;
 	}
 
+	if (provider != NULL) {
+		provider->enabled = FALSE;
+	}
+
+#if defined(ENABLE_PKCS11H_THREADING)
+free1:
+	for (
+		current_session = _g_pkcs11h_data->sessions;
+		current_session != NULL;
+		current_session = current_session->next
+	) {
+		_pkcs11h_threading_mutexRelease (&current_session->mutex);
+	}
+
+	if (has_mutex_cache) {
+		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.cache);
+		has_mutex_cache = FALSE;
+	}
+	if (has_mutex_session) {
+		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.session);
+		has_mutex_session = FALSE;
+	}
+	if (has_mutex_global) {
+		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
+		has_mutex_global = FALSE;
+	}
+
+	if (lock_rv != CKR_OK) {
+		rv = lock_rv;
+		goto cleanup;
+	}
+#endif
+
 	if (provider == NULL) {
 		rv = CKR_OBJECT_HANDLE_INVALID;
 		goto cleanup;
 	}
 
-	provider->enabled = FALSE;
 	provider->reference[0] = '\0';
 
 	if (provider->should_finalize) {
@@ -968,29 +995,6 @@ pkcs11h_removeProvider (
 
 cleanup:
 
-#if defined(ENABLE_PKCS11H_THREADING)
-	for (
-		current_session = _g_pkcs11h_data->sessions;
-		current_session != NULL;
-		current_session = current_session->next
-	) {
-		_pkcs11h_threading_mutexRelease (&current_session->mutex);
-	}
-
-	if (has_mutex_cache) {
-		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.cache);
-		has_mutex_cache = FALSE;
-	}
-	if (has_mutex_session) {
-		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.session);
-		has_mutex_session = FALSE;
-	}
-	if (has_mutex_global) {
-		_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
-		has_mutex_global = FALSE;
-	}
-#endif
-
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
 		"PKCS#11: pkcs11h_removeProvider return rv=%lu-'%s'",
@@ -1009,7 +1013,9 @@ pkcs11h_forkFixup (void) {
 #if defined(ENABLE_PKCS11H_THREADING)
 	return CKR_OK;
 #else
-	return __pkcs11h_forkFixup (TRUE);
+	if (_g_pkcs11h_data->safefork) {
+		return __pkcs11h_forkFixup ();
+	}
 #endif
 #endif
 }
@@ -1276,11 +1282,12 @@ __pkcs11h_threading_atfork_child (void) {
 	if (_g_pkcs11h_data != NULL && _g_pkcs11h_data->initialized) {
 		_pkcs1h_threading_mutexReleaseAll ();
 		if (_g_pkcs11h_data->safefork) {
-			__pkcs11h_forkFixup (TRUE);
-		}
-		else {
-			__pkcs11h_forkFixup (FALSE);
-			pkcs11h_terminate ();
+			static PKCS11H_BOOL in_forkfixup = FALSE;
+			if (!in_forkfixup) {
+				in_forkfixup = TRUE;
+				__pkcs11h_forkFixup ();
+				in_forkfixup = FALSE;
+			}
 		}
 	}
 }
@@ -1289,37 +1296,20 @@ __pkcs11h_threading_atfork_child (void) {
 
 static
 CK_RV
-__pkcs11h_forkFixup (
-	IN const PKCS11H_BOOL activate_slotevent
-) {
-#if defined(ENABLE_PKCS11H_THREADING)
-	PKCS11H_BOOL mutex_locked = FALSE;
-#endif
+__pkcs11h_forkFixup () {
 #if defined(ENABLE_PKCS11H_DEBUG)
 	pid_t mypid = getpid ();
 #endif
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: __pkcs11h_forkFixup entry pid=%d, activate_slotevent=%d",
-		mypid,
-		activate_slotevent ? 1 : 0
+		"PKCS#11: __pkcs11h_forkFixup entry pid=%d",
+		mypid
 	);
 
-#if !defined(ENABLE_PKCS11H_SLOTEVENT)
-	(void)activate_slotevent;
-#endif
-
 	if (_g_pkcs11h_data != NULL && _g_pkcs11h_data->initialized) {
 		_pkcs11h_provider_t current;
 
-#if defined(ENABLE_PKCS11H_THREADING)
-		if (_pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global) != CKR_OK) {
-			goto cleanup;
-		}
-		mutex_locked = TRUE;
-#endif
-
 		for (
 			current = _g_pkcs11h_data->providers;
 			current != NULL;
@@ -1336,22 +1326,10 @@ __pkcs11h_forkFixup (
 			 */
 			if (_g_pkcs11h_data->slotevent.initialized) {
 				_pkcs11h_slotevent_terminate_force ();
-
-				if (activate_slotevent) {
-					_pkcs11h_slotevent_init ();
-				}
+				_pkcs11h_slotevent_init ();
 			}
 #endif
 		}
-
-#if defined(ENABLE_PKCS11H_THREADING)
-	cleanup:
-
-		if (mutex_locked) {
-			_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
-			mutex_locked = FALSE;
-		}
-#endif
 	}
 
 	_PKCS11H_DEBUG (
